from typing import Dict
from .base import Transport
from arkitekt.config import TransportProtocol
import logging

logger = logging.getLogger(__name__)

class TransportRegistry:

    def __init__(self) -> None:
        self.registeredTransports: Dict[str, Transport] = {}

    def register_transport(self, protocol: TransportProtocol,  transport: Transport):
        self.registeredTransports[protocol] = transport

    def get_transport_for_protocol(self,protocol):
        return self.registeredTransports[protocol]

    
def register_transport(protocol: TransportProtocol):

    def real_decorator(transport):

        assert issubclass(transport, Transport), "Transport must subclass Transport"
        logger.info(f"Registerying Transport {transport} for {protocol}")
        get_current_transport_registry().register_transport(protocol, transport)
        return transport

    return real_decorator



GRANT_REGISTRY = None


def get_current_transport_registry(with_defaults=True):
    global GRANT_REGISTRY
    if not GRANT_REGISTRY:
        GRANT_REGISTRY = TransportRegistry()
        if with_defaults:
            import arkitekt.transport.websocket 


    return GRANT_REGISTRY


from arkitekt.schema.widgets import Widget
from arkitekt.schema.ports import Port
from typing import List

class NoWidgetRegisteredError(Exception):
    pass


class WidgetRegistry:

    def __init__(self) -> None:
        self.portTypeWidgetTypeQTWidgetMap = {}


    def get_widget_class(self, portType, widgetType):
        portType = portType.__class__.__name__ if isinstance(portType, Port) else portType
        widgetType = widgetType.typename if isinstance(widgetType, Widget) else widgetType
        try:
            return self.portTypeWidgetTypeQTWidgetMap[portType][widgetType]
        except KeyError as e:
            raise NoWidgetRegisteredError(f"No widget registered for {portType} and {widgetType}") from e


    def register_defaults(self):
        from arkitekt.ui.qtwidgets.qtlistsearchwidget import QTListSearchWidget
        from arkitekt.ui.qtwidgets.qtsliderwidget import QTSliderWidget

    def register(self, list: List):

        def register(widget_class):
            for pair in list:
                portType, widgetType = pair
                try:
                    portType = portType.__name__ if issubclass(portType, Port) else portType
                except:
                    portType = portType
                try:
                    widgetType = widgetType.__name__ if issubclass(widgetType, Widget) else widgetType
                except:
                    widgetType = widgetType
                self.portTypeWidgetTypeQTWidgetMap.setdefault(portType, {})[widgetType] = widget_class
            
            return widget_class


        return register


WIDGET_REGISTRY = None

def get_widget_registry(register_defaults=True):
    global WIDGET_REGISTRY
    if WIDGET_REGISTRY is None:
        WIDGET_REGISTRY = WidgetRegistry()
        if register_defaults: WIDGET_REGISTRY.register_defaults()

    return WIDGET_REGISTRY



from typing import Dict
from arkitekt.actors.actify import actify, define

from arkitekt.packers.transpilers.base import Transpiler
from arkitekt.schema.node import Node
from .app import AppAgent

class ScriptAgent(AppAgent):

    def register(self, *args, widgets={}, transpilers: Dict[str, Transpiler] = None, on_provide = None, on_unprovide = None, **params):

        def real_decorator(function):
            # Simple bypass for now
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            
            defined_actor = actify(function, on_provide=on_provide, on_unprovide=on_unprovide, **params)

            if len(args) == 0:
                defined_node = define(function=function, widgets=widgets)
                self.templatedNewNodes.append((defined_node, defined_actor, params))

            if len(args) == 1:
                if isinstance(args[0], str):
                    self.templatedUnqueriedNodes.append(({"q": args[0]}, defined_actor, params))

                if isinstance(args[0], Node):
                    self.templatedNodes.append((args[0], defined_actor, params))

                # We are registering this as a template

            return wrapped_function

    
        return real_decorator


from arkitekt.config import ArkitektConfig
from herre.wards.graphql import GraphQLWardConfig, ParsedQuery, GraphQLWard
from herre.herre import  get_current_herre




class ArkitektWard(GraphQLWard):
    configClass = ArkitektConfig

    class Meta:
        key = "arkitekt"


    async def negotiate(self):
        from arkitekt.schema.negotiation import Transcript
        transcript_query = await self.arun(ParsedQuery("""mutation Negotiate {
            negotiate {
                postman {
                    type
                    kwargs
                }
            } 
        }"""))
        return Transcript(**transcript_query["negotiate"])



from ..exception import ExceptionDataModel
from pydantic.main import BaseModel
from ....messages.types import  ASSIGN_CRITICAL
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = ASSIGN_CRITICAL
    extensions: Optional[MetaExtensionsModel]

class AssignCriticalMessage(MessageModel):
    data: ExceptionDataModel
    meta: MetaModel

    @classmethod
    def from_critical(cls,reference, exception, extensions={}):
        return cls(data={
            "message": str(exception),
            "type": exception.__class__,
            }, meta={"extensions": extensions, "reference":reference})


import os


IMAGES_PATH = os.path.dirname(os.path.realpath(__file__))

def get_image_path(filename, darkMode=False):
    base_folder =  os.path.join(IMAGES_PATH,"darkmode") if darkMode else os.path.join(IMAGES_PATH, "lightmode") 
    return os.path.join(base_folder,filename)





import asyncio
import sys
import time
import logging
from watchdog.observers import Observer
from watchdog.events import FileModifiedEvent, FileSystemEventHandler, LoggingEventHandler
import os
import subprocess
import janus
import threading
import os
import signal


class QueueHandler(FileSystemEventHandler):

    def __init__(self, *args, sync_q=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = sync_q

    def on_any_event(self, event):
        self.queue.put(event)
        self.queue.join()



def watcher(path, queue, event):
    event_handler = QueueHandler(sync_q=queue)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    observer.start()

    while True:
        time.sleep(1)


async def buffered_queue(queue, timeout= 4):
    buffer = [None]

    async def iterator():
        while True:
            nana = await queue.get()
            queue.task_done()
            if isinstance(nana, FileModifiedEvent):
                if nana.src_path.lower().endswith((".py")):
                    buffer.append(nana)

    loop = asyncio.get_event_loop()      
    loop.create_task(iterator())
    while True:
        await asyncio.sleep(1)
        yield buffer[len(buffer) - 1]
        buffer = [None]


async def _read_stream(stream, cb):  
    while True:
        line = await stream.readline()
        if line:
            cb(line)
        else:
            break



async def run_subproc(cmd, outcb = print, errorcb = print):
    my_env = os.environ.copy()
    my_env["PYTHONUNBUFFERED"] = "1"

    try:
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env = my_env)

        await asyncio.wait([
        _read_stream(proc.stdout, outcb),
        _read_stream(proc.stderr, errorcb)
        ])
        await proc.wait()
        print("Program Failed. Please Correct Error!")
        return

    except asyncio.CancelledError as e:
        proc.kill()
        proc.terminate()
        await proc.wait()
        raise e



class Host():

    def __init__(self, path = None, run_py = "run.py", outcb=print, errorcb=print) -> None:
        self.script_path = os.path.join(path, run_py)
        self.interpreter_path = sys.executable
        self.process_task = None
        self.outcb = outcb
        self.errorcb = errorcb


    async def restart(self):
        loop = asyncio.get_event_loop()
        if self.process_task:
            print("Cancelling our Initial Task")
            self.process_task.cancel()
            try:
                await self.process_task
            except asyncio.CancelledError as e:
                print("Process Completely Cancelled")

        print("Starting new Task")
        self.process_task = loop.create_task(run_subproc([self.interpreter_path, self.script_path], outcb=self.outcb, errorcb=self.errorcb))


    async def run(self, queue):
        await self.restart()

        async for event in buffered_queue(queue, timeout=1):
            if event:
                await self.restart()
        


async def watch_directory_and_restart(path = "watch", run_py = "run.py"):

    cancel_event = threading.Event()
    jqueue = janus.Queue()
    loop = asyncio.get_running_loop()
    fut = loop.run_in_executor(None, watcher, path, jqueue.sync_q, cancel_event)

    host = Host(path=path, run_py=run_py)
    await host.run(jqueue.async_q)
    await fut

    jqueue.close()
    await jqueue.wait_closed()




from arkitekt.monitor.monitor import get_current_monitor
from herre.console.context import get_current_console
from uuid import uuid4
from arkitekt.packers.utils import expand_outputs, shrink_inputs
from asyncio.futures import Future
from arkitekt.messages.postman.log import LogLevel
from contextvars import Context
from arkitekt.messages.postman.reserve.reserve_transition import ReserveState
from arkitekt.contracts.exceptions import AssignmentException
from arkitekt.messages import *
from arkitekt.schema.enums import NodeType
from arkitekt.messages.postman.reserve.params import ReserveParams
from arkitekt.postman import Postman
from rich.table import Table
from rich.panel import Panel
from arkitekt.monitor import Monitor, current_monitor
from arkitekt.registry import get_current_postman
import asyncio
import logging
from herre.herre import Herre, get_current_herre

from koil.koil import Koil, get_current_koil
from koil.loop import koil, koil_gen

logger = logging.getLogger(__name__)

class UnknownMessageError(Exception):
    pass


class ReservationError(Exception):
    pass


class CouldNotReserveError(ReservationError):
    pass


class IncorrectStateForAssignation(ReservationError):
    pass


def build_reserve_message(reference, node_id: str = None, template_id: str= None, provision: str= None, params_dict: dict = {}, with_log=False, context=None):
    assert reference is not None, "Must have a reference"
    assert node_id is not None or template_id is not None, "Please provide either a node_id or template_id"

    data={
                                "node": node_id, 
                                "template": template_id, 
                                "provision": provision,
                                "params": params_dict,
    }
    meta={
        "reference": reference,
        "extensions": {
            "with_progress": with_log,
        }
    }

    if context:
        meta = {**meta, "context": context}
        return BouncedReserveMessage(data=data, meta=meta)

    else:
        return ReserveMessage(data=data, meta=meta)

def build_unreserve_messsage(reference, reservation, with_log=False, context=None):
    assert reference is not None, "Must have a reference"
    data= {
                                        "reservation": reservation
    }
    
    meta={
                                    "reference": reference, 
                                    "extensions": {
                                        "with_progress": with_log
                                    }
    }

    if context:
        meta = {**meta, "context": context}
        return BouncedUnreserveMessage(data=data, meta=meta)

    else:
        return UnreserveMessage(data=data, meta=meta)

def build_assign_message(reference, reservation, args, kwargs, with_log=False, context=None, persist=False):
    assert reference is not None, "Must have a reference"

    data = {
                                    "reservation": reservation,
                                    "args": args, 
                                    "kwargs": kwargs,
    }

    meta = {
                                    "reference": reference, 
                                    "extensions": {
                                        "with_progress": with_log,
                                        "persist": persist
                                    }
    }

    if context:
        meta = {**meta, "context": context}
        return BouncedAssignMessage(data=data, meta=meta)

    else:
        return AssignMessage(data=data, meta=meta)

def build_unassign_messsage(reference, assignation, with_log=False, context=None, persist=False):
    assert reference is not None, "Must have a reference"
    data= {
                                        "assignation": assignation
    }
                                    
    meta={
                                    "reference": reference, 
                                    "extensions": {
                                        "with_progress": with_log,
                                        "persist": persist
                                    }
    }

    if context:
        meta = {**meta, "context": context}
        return BouncedUnassignMessage(data=data, meta=meta)

    else:
        return UnassignMessage(data=data, meta=meta)


class Reservation:

    def __init__(self, node,
        reference: str = None,
        provision: str = None, 
        monitor: Monitor = None,
        ignore_node_exceptions=False,
        transition_hook=None,
        with_log=False,
        enter_on=[ReserveState.ACTIVE], 
        exit_on=[ReserveState.ERROR, ReserveState.CANCELLED, ReserveState.CRITICAL],
        context: Context =None,
        koil: Koil = None,
        herre: Herre = None,
        postman: Postman = None,
         **params) -> None:

        self.monitor: Monitor = monitor or get_current_monitor()
        self.panel = self.monitor.create_reservation_panel(self) if self.monitor else None


        self.console = get_current_console()
        self.herre = herre or get_current_herre()
        self.koil = koil or get_current_koil()
        self.loop = self.koil.loop
        self.postman = postman or get_current_postman(force_creation=True)
        

        # Reservation Params
        self.reference = reference or str(uuid.uuid4())
        self.provision = provision
        self.node = node
        self.params = ReserveParams(**params)
        self.with_log = with_log or (self.monitor.log if self.monitor else None)


        self.context = context # with_bounced allows us forward bounced checks
        if self.context:
            assert "can_forward_bounce" in self.herre.grant.scopes, "In order to use with_bounced forwarding you need to have the can_forward_bounced scope"

        # Exception Mangement
        self.ignore_node_exceptions = ignore_node_exceptions
        self.critical_error = None

        # State management
        self.transition_hook = transition_hook
        assert self.transition_hook is None or asyncio.iscoroutinefunction(self.transition_hook), "Transition Hook must be either a coroutine or set to None"
        self.exit_states = exit_on
        self.enter_states = enter_on
        self.current_state = ReserveState.STARTING



    
    def log(self, message: str, level: LogLevel=LogLevel.DEBUG):
        """Logs a Message

        The Logged Message will be display on the Monitor if running inside a Monitor
        and send to the logging output.

        Args:
            message (str): The Message
            level (LogLevel, optional): The LogLevel. Defaults to LogLevel.DEBUG.
        """
        if self.panel:
            self.panel.log(message, level=level)
        logger.info(f"{level}: {message}")


    async def transition_state(self, message: ReserveTransitionMessage):
        # Once we acquire a reserved resource our contract (the inner part of the context can start)
        if self.transition_hook: await self.transition_hook(self, message.data.state)
        if message.data.state in self.exit_states:
            if self.enter_future.done():
                self.log(f"We have transitioned to a critical State {message.data.message}. Terminating on Next Call")
            else:
                self.log("Cancelling Reservation")
                raise Exception(message.data.message)

            if not self.is_closing: 
                self.log(f"Received Exitstate: {message.data.state}. Closing reservation at next assignment", level=LogLevel.CRITICAL)
            

        if message.data.state in self.enter_states:
            if self.enter_future.done():
                logger.info("We are already entered.")
            else:
                self.enter_future.set_result(message.meta.reference)



        self.old_state = self.current_state
        self.current_state = message.data.state
        self.log(f"[red] {self.old_state} > {self.current_state}: {message.data.message}", LogLevel.INFO)


    async def assign_async(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, with_log=True, context=None, raise_node_exceptions=True, **kwargs):
        logger.info(f"Assigning {args} {kwargs} ")
        assert self.node.type == NodeType.FUNCTION, "You cannot assign to a Generator Node, use the stream Method!"

        if self.current_state in self.exit_states:
            raise IncorrectStateForAssignation(f"Current State {self.current_state} is an Element of Exit States {self.exit_states}")
        

        
        shrinked_args, shrinked_kwargs = await shrink_inputs(self.node, *args, **kwargs) if not bypass_shrink else (args, kwargs)
        


        context = context or self.context
        assign_reference = str(uuid.uuid4())
        assign_message = build_assign_message(assign_reference, self.reference, shrinked_args, shrinked_kwargs, with_log=with_log, context=context)
        assignation_queue = await self.postman.stream_replies_to_queue(assign_message)

        try:
            while True:
                message = await assignation_queue.get()

                if self.current_state in self.exit_states:
                    raise IncorrectStateForAssignation(f"Current State {self.current_state} is an Element of Exit States {self.exit_states}")

                if isinstance(message, AssignReturnMessage):
                    return await expand_outputs(self.node, message.data.returns) if not bypass_expand else message.data.returns    

                if isinstance(message, AssignCancelledMessage):
                    raise AssignmentException(f"Assignment was cancelled from a different Agent: ID: {message.data.canceller}")

                if isinstance(message, AssignLogMessage):
                    self.log(message.data.message, message.data.level)
                    continue

                if isinstance(message, AssignReceivedMessage):
                    self.log(f"Received the Message on Provision {message.data.provision}", LogLevel.INFO)
                    continue

                if isinstance(message, AssignCriticalMessage):
                    if raise_node_exceptions:
                        raise AssignmentException(message.data.message)
                    else:
                        self.log(message.data.message, LogLevel.CRITICAL)

                if isinstance(message, AssignYieldsMessage):
                    raise AssignmentException("Received a Yield from a Node that should never yield! CRITICAL PROTOCOL EXCEPTION")

                raise UnknownMessageError(message)


        except asyncio.CancelledError as e:
            self.log("Assigment Required Cancellation", level=LogLevel.INFO)

            un_assign_reference = str(uuid.uuid4())
            unassign_message = build_unassign_messsage(un_assign_reference, assign_reference, context=context)
            

            self.log(f"Cancellation Condition {unassign_message}")
            await self.postman.transport.forward(unassign_message)

            while True:
                message = await assignation_queue.get()
                if isinstance(message, AssignCancelledMessage):

                    if message.data.canceller != un_assign_reference:
                        self.log("Canceller does not match our Cancellation Request, Race Condition?")
                    raise e

                else:
                    logger.info(f"Raced Condition on Cancellation Message, we received {message} before even though we wanted a cancellation. Ommiting this")



    async def stream_async(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, context=None, with_log=True, **kwargs):
        assert self.node.type == NodeType.GENERATOR, "You cannot stream a Function Node, use the assign Method!"

        if self.current_state in self.exit_states:
            raise IncorrectStateForAssignation(f"Current State {self.current_state} is an Element of Exit States {self.exit_states}")

        shrinked_args, shrinked_kwargs = await shrink_inputs(self.node, *args, **kwargs) if not bypass_shrink else (args, kwargs)
        
        context = context or self.context
        assign_reference = str(uuid.uuid4())
        assign_message = build_assign_message(assign_reference, self.reference, shrinked_args, shrinked_kwargs, with_log=with_log, context=context)
        assignation_queue = await self.postman.stream_replies_to_queue(assign_message)

        try:
            while True:
                message = await assignation_queue.get()

                if isinstance(message, AssignYieldsMessage):
                    yield await expand_outputs(self.node, message.data.returns) if not bypass_expand else message.data.returns    

                elif isinstance(message, AssignDoneMessage):   
                    break

                elif isinstance(message, AssignCancelledMessage):
                    raise AssignmentException(f"Assignment was cancelled from a different Agent: ID: {message.data.canceller}")

                elif isinstance(message, AssignLogMessage):
                    self.log(message.data.message, message.data.level)

                elif isinstance(message, AssignReceivedMessage):
                    self.log(f"Received the Message on Provision {message.data.provision}", LogLevel.INFO)
                    continue

                elif isinstance(message, AssignCriticalMessage):
                    raise AssignmentException(message.data.message)

                elif isinstance(message, AssignReturnMessage):
                    raise AssignmentException("Received a Return from a Node that should never return! CRITICAL PROTOCOL EXCEPTION")

                else:
                    raise UnknownMessageError(message)


        except asyncio.CancelledError as e:
            self.log("Assigment Required Cancellation", level=LogLevel.INFO)

            un_assign_reference = str(uuid.uuid4())
            unassign_message = build_unassign_messsage(un_assign_reference, assign_reference, context=context)
            
            await self.postman.transport.forward(unassign_message)


            while True:
                message = await assignation_queue.get()

                if isinstance(message, AssignCancelledMessage):

                    if message.data.canceller != un_assign_reference:
                        self.log("Canceller does not match our Cancellation Request, Race Condition?")
                    raise e

                else:
                    logger.info(f"Wrong message for cancellation {message}")



    async def stream_worker(self):

        reserve_message = build_reserve_message(reference=self.reference, node_id=self.node.id, params_dict=self.params.dict())
        reservation_queue = await self.postman.stream_replies_to_queue(reserve_message)

        try:
            while True: 
                message = await reservation_queue.get()

                if isinstance(message, ReserveLogMessage):
                    self.log(message.data.message, message.data.level)


                elif isinstance(message, ProvideTransitionMessage):
                    self.log(message.data.message, message.data.state)
                

                elif isinstance(message, ReserveTransitionMessage):
                    
                        await self.transition_state(message)

                else:
                    self.enter_future.set_exception(UnknownMessageError("Received Unnown Message"))

                reservation_queue.task_done()

        except asyncio.CancelledError as e:

            unreserve_reference = str(uuid.uuid4())
            unreserve_message = build_unreserve_messsage(unreserve_reference, self.reference,  context=self.context)
            
            await self.postman.transport.forward(unreserve_message)


            while True:
                message = await reservation_queue.get()

                if isinstance(message, ReserveTransitionMessage):
                    if self.transition_hook: await self.transition_hook(self, message.data.state)
                    if message.data.state in [ReserveState.CANCELLED]:
                        raise e
                    else:
                        self.console.log(f"[red] Received Completely Different State {message.data.state}")

                else:
                    self.console.log(f"[red] Received Completely Different Message {message}")

        
        except Exception as e: 
            logger.exception(e)
            if not self.enter_future.done():
                self.enter_future.set_exception(e)
            raise e


    async def cancel(self):
        self.is_closing = True
        self.stream_task.cancel()
        try:
            await self.stream_task
        except asyncio.CancelledError as e:
            pass

    async def start(self):
        return await self.__aenter__()

    async def end(self, timeout=3):
        #TODO: implement timeout
        await self.cancel()

    async def __aenter__(self):
        if self.panel: self.panel.start()

        close_postman = False
        # Check connection level
        if not self.postman.connected:
            close_postman = True
            await self.postman.aconnect()

        self.is_closing = False
        
        self.enter_future = self.loop.create_future()
        self.stream_task = self.loop.create_task(self.stream_worker())

        try:
            self.enter_state = await self.enter_future
            return self

        except Exception as e:
            logger.exception(e)
            if close_postman: 
                logger.debug("Waiting to disconnect the postman that we previously connected!")  
                await self.postman.adisconnect()
            await self.cancel()

            raise CouldNotReserveError(f"Could not Reserve Reservation {self.reference} for Node {self.node}") from e


    async def __aexit__(self, type, value, traceback):

        await self.cancel()
        if self.panel: self.panel.end()
        
        if type is not None:
            if issubclass(type, asyncio.CancelledError): 
                raise type(value).with_traceback(traceback)

            if issubclass(type, Exception):
                raise type(value).with_traceback(traceback)


            raise type(value).with_traceback(traceback)


    def stream(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, **kwargs):
        return koil_gen(self.stream_async(*args, bypass_shrink=bypass_shrink, bypass_expand=bypass_expand, persist=persist, **kwargs))

    def assign(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, **kwargs):
        return koil(self.assign_async(*args, bypass_shrink=bypass_shrink, bypass_expand=bypass_expand, persist=persist, **kwargs))

    def __enter__(self):
        return koil(self.__aenter__())

    def __exit__(self,*args, **kwargs):
        return koil(self.__aexit__(*args, **kwargs))





    




       


from ....messages.generics import Context
from pydantic.main import BaseModel
from .params import ReserveParams
from ....messages.types import  BOUNCED_FORWARDED_RESERVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional

class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = BOUNCED_FORWARDED_RESERVE
    extensions: Optional[MetaExtensionsModel]
    context: Context

class DataModel(MessageDataModel):
    node: Optional[str] #TODO: Maybe not optional
    template: Optional[str]
    params: Optional[ReserveParams]
    provision: str

class BouncedForwardedReserveMessage(MessageModel):
    data: DataModel
    meta: MetaModel




from typing import Dict
from arkitekt.actors.actify import actify, define
from arkitekt.agents.app import AppAgent
from arkitekt.agents.qt.actor import QtActor
from qtpy.QtCore import QObject, Signal
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from arkitekt.agents.base import Agent, AgentException
import logging
import uuid

logger = logging.getLogger(__name__)


class AgentSignals(QObject):
    provide = Signal(BouncedProvideMessage)
    unprovide = Signal(BouncedUnprovideMessage)
    provide_transition = Signal(ProvideTransitionMessage)


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class QtAgent(AppAgent, QObject):
    provide_signal = Signal(bool)
    provision_signal = Signal(BouncedProvideMessage)
    unprovision_signal = Signal(BouncedUnprovideMessage)

    ACTOR_PENDING_MESSAGE = "Actor is Pending"

    def __init__(self,*args, strict=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.strict = strict
        self.assignFutures = {}
        self.provideFutures = {}
        self.unprovideFutures = {}
        self.appWorkers = {}


    async def on_bounced_provide(self, message: BouncedProvideMessage):
        self.provision_signal.emit(message)
        return await super().on_bounced_provide(message)


    async def on_bounced_unprovide(self, message: BouncedUnprovideMessage):
        self.unprovision_signal.emit(message)
        return await super().on_bounced_provide(message)


    def register_ui(self, function_query_or_node, widgets={}, transpilers: Dict[str, Transpiler] = None, on_provide = None, on_unprovide = None, on_assign = None, timeout=500, **params) -> QtActor:

        # Simple bypass for now
        defined_actor = QtActor(qt_assign=on_assign, qt_on_provide=on_provide, qt_on_unprovide=on_unprovide, loop=self.loop, timeout=timeout, **params)
        
        if isinstance(function_query_or_node, str):
            self.templatedUnqueriedNodes.append(({"q": function_query_or_node}, defined_actor, params))
           
        if isinstance(function_query_or_node, Node):
            self.templatedNodes.append((function_query_or_node, defined_actor, params))

        else:
            defined_node = define(function=function_query_or_node, widgets=widgets)
            self.templatedNewNodes.append((defined_node, defined_actor, params))

        return defined_actor


    def register_side(self, *args, widgets={}, transpilers: Dict[str, Transpiler] = None, on_provide = None, on_unprovide = None, on_assign = None, timeout=500, **params):

        # Simple bypass for now

        if len(args) == 0: raise NotImplementedError("Please provide either a function to create a node or a function and a node or query as arguments")


        defined_actor = actify(args[0], on_provide=on_provide, on_unprovide=on_unprovide, **params)


        if len(args) == 1: 
            new_node = define(function=args[0], widgets=widgets)
            self.templatedNewNodes.append((new_node, defined_actor, params))
           
        
        if len(args) == 2:
            query_or_node = args[1]    

            if isinstance(query_or_node, str):
                self.templatedUnqueriedNodes.append(({"q": query_or_node}, defined_actor, params))
           
            if isinstance(query_or_node, Node):
                self.templatedNodes.append((query_or_node, defined_actor, params))
        
        return defined_actor


    async def approve_nodes_and_templates(self):
        await super().approve_nodes_and_templates()
        self.provide_signal.emit(True)
        return 

    async def aprovide(self):
        try:
            await super().aprovide()
        except asyncio.CancelledError as e:
            self.provide_signal.emit(False)
            raise e


    
   

    


from pydantic.main import BaseModel
from arkitekt.config import TransportProtocol
from arkitekt.transport.base import Transport, TransportConfig
from arkitekt.messages.utils import expandToMessage
from arkitekt.messages.base import MessageMetaExtensionsModel, MessageModel
from arkitekt.transport.registry import register_transport
from herre.herre import get_current_herre
from enum import Enum
import websockets
import json
import asyncio
from arkitekt.legacy.utils import create_task
import logging

logger = logging.getLogger(__name__)


class WebsocketTransportConfig(BaseModel):
    host: str
    port: int
    secure: bool = False
    route: str

    @property
    def protocol(self):
        return "wss" if self.secure else "ws"


class ConnectionFailedError(Exception):
    pass

@register_transport(TransportProtocol.WEBSOCKET)
class WebsocketTransport(Transport):
    configClass = WebsocketTransportConfig
    config: WebsocketTransportConfig


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.retries = 5
        self.time_between_retries = 5
        self.connection_alive = False
        self.connection_dead = False


    async def aconnect(self):
        if not self.herre.logged_in:
            await self.herre.alogin()

        self.send_queue = asyncio.Queue()
        self.connection_task =  create_task(self.websocket_loop())


    async def adisconnect(self):
        self.connection_task.cancel()

        try:
            await self.connection_task
        except asyncio.CancelledError:
            logger.info(f"Websocket Transport {self} succesfully disconnected")
    


    async def websocket_loop(self, retry=0):
        send_task = None
        receive_task = None

        assert retry < self.retries, "Exceeded number of retries! Postman is disconnected"
        try:
            try:
                async with websockets.connect(f"{self.config.protocol}://{self.config.host}:{self.config.port}{self.config.route}/?token={self.herre.state.access_token}") as client:

                    send_task = create_task(self.sending(client))
                    receive_task = create_task(self.receiving(client))

                    self.connection_alive = True
                    self.connection_dead = False
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self.connection_alive = True

                    for task in pending:
                        task.cancel()

                    raise ConnectionFailedError("Connection failed")

            except Exception as e:
                raise ConnectionFailedError from e

        except ConnectionFailedError as e:
            logger.error("Connection to failed Retrying")
            await asyncio.sleep(self.time_between_retries)
            await self.connection_task(retry=retry + 1)

        except AssertionError as e:
            logger.error("Connection failed Definetly. Postman will fail on next call!")
            self.connection_dead = False

        except asyncio.CancelledError as e:
            logger.info("Got Canceleld")
            if send_task and receive_task:
                 send_task.cancel()
                 receive_task.cancel()

            cancellation = await asyncio.gather(send_task, receive_task, return_exceptions=True)
            raise e


    async def sending(self, client):
        try:
            while True:
                message = await self.send_queue.get()
                logger.debug("Websocket: >>>>>> " + message)
                await client.send(message)
                self.send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.debug("Sending Task sucessfully Cancelled")


    async def receiving(self, client):
        try:
            async for message in client:
                logger.debug("Websocket: <<<<<<< " + message)
                message = expandToMessage(json.loads(message))
                await self.broadcast(message)
        except asyncio.CancelledError as e:
            logger.debug("Receiving Task sucessfully Cancelled")


    async def forward(self, message: MessageModel):
        assert not self.connection_dead, "Connection is definetly dead. Retries have been exceeded. Error"
        await self.send_queue.put(json.dumps(message.dict()))
        


    


from arkitekt.ui.registry import get_widget_registry
from typing import Type
from qtpy import QtWidgets
from arkitekt.ui.qtwidgets.base import UIPortMixin
from arkitekt.schema.ports import *
from herre.wards.graphql import ParsedQuery
from herre.access.model import GraphQLModel

registry = get_widget_registry()

class QueryOption(BaseModel):
    label: str
    value: str


@registry.register(
    [(ListArgPort, SearchWidget ),(ListKwargPort, SearchWidget)]
)
class QTListSearchWidget(QtWidgets.QWidget, UIPortMixin):
    port: ListArgPort

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.build_ui()
        self.selected_value = None
        self.child: StructureArgPort = self.port.child
        self.structure = get_packer_registry().get_structure(self.child.identifier)
        assert issubclass(self.structure, GraphQLModel), "Selected Port is not a GraphQL Model. Search widget will not work"


    def build_ui(self):
        self.layout = QtWidgets.QVBoxLayout()


        self.searchBox = QtWidgets.QLineEdit("")
        self.searchBox.textChanged.connect(self.update_list)

        self.listWidget = QtWidgets.QListWidget()
        self.listWidget.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection 
        )


        self.layout.addWidget(self.searchBox)
        self.layout.addWidget(self.listWidget)
        self.layout.addWidget(QtWidgets.QLabel(self.port.description))

        self.listWidget.itemClicked.connect(self.update_value)

        self.setLayout(self.layout)


    def update_list(self, searchString):
        if len(searchString) > 1:

            results = self.structure.objects.ward._run_sync(ParsedQuery(self.port.widget.query), {"search": searchString})
            self.options = [QueryOption(**values) for values in results["options"]]

            self.listWidget.clear()
            for option in self.options:
                item = QtWidgets.QListWidgetItem(option.label)
                self.listWidget.addItem(item)

    
    def update_value(self):
        items = self.listWidget.selectedIndexes()
        self.selected_value = [self.structure(id=self.options[index.row()].value) for index in items]


    def get_current_value(self):
        return self.selected_value 







from arkitekt.actors.functional import FunctionalFuncActor
from herre.herre import get_current_herre
from arkitekt.threadvars import get_current_assign
from qtpy.QtCore import QObject, Signal
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.provide.provide_critical import ProvideCriticalMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from herre.wards.base import WardException
from arkitekt.actors.actify import actify, define
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.base import MessageDataModel, MessageModel
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.template import Template
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.agents.base import Agent, AgentException, parse_params
import logging
from herre.console import get_current_console
import uuid
from koil import get_current_koil
from koil.koil import Koil
from koil.qt import FutureWrapper, TimeoutFutureWrapper

logger = logging.getLogger(__name__)





class ActorSignals(QObject):

        def __init__(self, timeout=300) -> None:
            self.assign = TimeoutFutureWrapper(timeout=timeout)
            self.on_provide = TimeoutFutureWrapper(timeout=timeout, pass_through=True)
            self.on_unprovide = TimeoutFutureWrapper(timeout=timeout, pass_through=True)


class QtActor(FunctionalFuncActor, QObject):

    def __init__(self, *args,  qt_assign=None, loop= None, qt_on_provide = None, qt_on_unprovide = None, timeout=300, koil: Koil = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.koil = koil or get_current_koil()
        self.loop = self.koil.loop

        self.qt_assign = qt_assign
        self.qt_on_provide = qt_on_provide
        self.qt_on_unprovide = qt_on_unprovide

        self.signals = ActorSignals()

        if self.qt_assign is not None:
            self.signals.assign.call.connect(self.assign_function_in_qtloop) # insignal will be set dymaically by qt agent

        if self.qt_on_provide is not None:
            self.signals.on_provide.connect(self.provide_function_in_qtloop)

        if self.qt_on_unprovide is not None:
            self.signals.on_unprovide.connect(self.unprovide_function_in_qtloop)

        

    def provide_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_on_provide(*args, **kwargs) if self.qt_on_provide else None
            self.signals.on_provide.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.on_provide.reject.emit(reference, e)

    def unprovide_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_on_unprovide(*args, **kwargs) if self.qt_on_unprovide else None
            self.signals.on_unprovide.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.on_unprovide.reject.emit(reference, e)


    def assign_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_assign(*args, **kwargs)
            self.signals.assign.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.assign.reject.emit(reference, e)


    async def assign(self, *args, **kwargs):
        await self.signals.assign.acall(*args, **kwargs)

    async def on_provide(self, *args, **kwargs):
        await self.signals.on_provide.acall(*args, **kwargs)

    async def on_unprovide(self, *args, **kwargs):
        await self.signals.on_unprovide.acall(*args, **kwargs)


    

